/* bootld.c: H8 bootload kernel  */
/*
** Copyright (C) 2000,2001 Martin Glunz <mg-home@gmx.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "h83048f.h"
#include "flash.h"
#include "bootld.h"
#include "timed.h"

#define PROG_ATTEMPT           6
#define ERASE_ATTEMPT_LIMIT    6

unsigned char status;

void verify(char *addr, unsigned long len, char *buf)
{
  status=0xfe;
  while (len--) if (*addr++!=*buf++) return;
  status=0;
}

void program(char *addr, unsigned long len, char *buf)
{
  unsigned char attempt, done, temp;
  unsigned short delay,j;
  status=0;

  FLMCR = VPPE_BIT;
  SETTLE_DELAY();
  EBR = 0xffff;

  while (len) {
    attempt=done=0;
    delay=1;
    while (!done) {
      *addr = *buf;
      FLMCR |= PROG_BIT;
      for (j=delay; j!= 0; j--) PROG_DELAY();
      FLMCR &= ~PROG_BIT;
      FLMCR |= PROGV_BIT;    
      PROG_VERIFY_DELAY();    
      temp  = *addr;                   
      FLMCR &= ~PROGV_BIT;    

      if (temp != *buf)   {
        attempt++; 
        delay = delay *2 ;      
        if(attempt > PROG_ATTEMPT) {
          done = 1;
          status=0xff;
	};
      } else done=1; 
    };
    addr++; buf++; len--;
  };
  EBR = 0 ;
  FLMCR = 0 ;                            
}   
    
void pre_write(char *addr, unsigned long len)
{
  unsigned char attempt, done, temp;
  unsigned short delay,j;
  status=0;
  FLMCR = VPPE_BIT;
  SETTLE_DELAY();
  EBR = 0xffff;
  while (len) {
    attempt=done=0;
    delay=1;
    while (!done) {
      *addr = 0x00;
      FLMCR |= PROG_BIT;
      for (j=delay; j!= 0; j--) PROG_DELAY();
      FLMCR &= ~PROG_BIT;
      PROG_VERIFY_DELAY();
      temp  = *addr;
      if (temp != 0x00)   {
	attempt++;
	delay = delay *2 ;
	if(attempt > PROG_ATTEMPT) {
	  done = 1;
	  status=0xff;
	};
      } else done=1;
    };
    addr++; len--;
  };
  EBR = 0;
  FLMCR = 0;
}

void erase_flash(short blocks, unsigned char delay)
{
  FLMCR = VPPE_BIT;               /* activate VPPE */
  SETTLE_DELAY();                                 /* activate a delay */
  EBR = blocks;                   /* set up the blocks to erase */
  SETTLE_DELAY();                                 /* activate a delay */
  FLMCR = (VPPE_BIT + ERASE_BIT); /* activate programming */
  for(;delay!=0; delay--) {
    ERASE_DELAY();                              /* activate a delay  for the specified amount of times */
  }
  EBR = 0 ;
  FLMCR = 0;                     /* de-activate programming */
}

int erase_verify(char *start, unsigned long len, short blocks)
{
  FLMCR = VPPE_BIT + ERASEV_BIT;  /* activate VPPE */
  SETTLE_DELAY();
  EBR = blocks;   /* set the EBR word */
  SETTLE_DELAY();
  while (len) {
    *start=0xff;
    ERASE_VERIFY_DELAY();
    if(*start != (unsigned char)0xff){
      EBR = 0;
      FLMCR = 0;
      return -1;
    };
    start++;
    len--;
  };
  EBR = 0;
  FLMCR = 0;
  return 0;
}

void erase(char *addr, unsigned long len)
{
  int delay, attempt_counter;
  status=0; delay=1;
  attempt_counter=0;
  pre_write(0, 0x20000);
  do {
    erase_flash(0xffff, delay);
    if (erase_verify(0, 0x20000, 0xffff)==0) {
      return;
    }
    delay = delay << 1;
    attempt_counter++;
  } while (attempt_counter < ERASE_ATTEMPT_LIMIT);
  status=0xff;
}

void reset(void) 
{
  WDT_WTCSR=0xA567; // Starte den Watchdog mit 1/4096 CLK      
  while(1);
}

               
void txchar(unsigned char data)
{
  while((SCI1_SSR & 0x80) == 0);
  SCI1_TDR = data;
  SCI1_SSR &= ~0x80;  
}
                            
unsigned char rxchar(void)
{
  while((SCI1_SSR & 0x40) == 0);
  SCI1_SSR &= ~0x40;  
  return SCI1_RDR;
}
       
unsigned char *c;
short len,i;
long addr;
unsigned char buf[BUFLEN];
	                    
void main(void)
{
  SCI1_SMR=0x00;
  SCI1_SCR=0x30;
  while (1) {
    switch (rxchar()) {
      case CMD_ID :
        txchar('m'); txchar('g'); txchar(SCI1_BRR);
        break;
      case CMD_BAUD :
        SCI1_BRR = rxchar();
        break;
      case CMD_ADDR :
        c=(void *)&addr; i=sizeof(addr);
	goto rxdt;
      case CMD_LEN :
        c=(void *)&len; i=sizeof(len);
        goto rxdt;
      case CMD_BUF :
        c=buf; i=len;
	rxdt:
        while (i--) *c++=rxchar();
        break;
      case CMD_PGM :
        program((void *)addr, len, buf);
        txchar(status);
	addr+=len;
        break;
      case CMD_CPY :
        memcpy(addr,buf,len);
	txchar(0);
	addr+=len;
	break;
      case CMD_VRFY :
        verify((void *)addr, len, buf);
	txchar(status);
	addr+=len;
        break;
      case CMD_RST :
        reset();
	break;
    };
  };
}
