/* h8boot.c: H8 boot loader for Linux */
/*
** Copyright (C) 2000,2001 Martin Glunz <mg-home@gmx.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <termios.h>
#include <time.h>
#include "bootld.h"
#include "bootld.bin.h"

#define CONNTRY 16
#define OVERLAP 16
#define WAITFORERASE 20000
#define INIBAUD B9600

int uart_fd;

int inituart(char *dev)
{
  struct termios newtio;
  int fd;
  fd = open(dev, O_RDWR | O_NOCTTY);
  if (fd < 0) { perror(dev); return -1; }
  newtio.c_cflag = INIBAUD | CS8 | CLOCAL | CREAD | HUPCL;
  newtio.c_iflag = IGNPAR;
  cfmakeraw(&newtio);
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd, TCSANOW, &newtio);
  return fd;
}

int setbaud(int fd, int baud)
{
  struct termios newtio;
  tcgetattr(fd, &newtio);
  cfsetispeed(&newtio,baud);
  cfsetospeed(&newtio,baud);
  tcsetattr(fd, TCSANOW, &newtio);
  return fd;
}

int select_fd(int fd)
{
  fd_set fds;
  struct timeval tm;
  tm.tv_sec=0;
  tm.tv_usec=0;
  FD_ZERO(&fds);
  FD_SET(fd,&fds);
  select(fd+1,&fds,0,0,&tm);
  return FD_ISSET(fd,&fds);
}

void udelay(int us)
{
  struct timespec req, rem;
  if (us>=1000000) {
    req.tv_sec=us/1000000;
    req.tv_nsec=(us % 1000000)*1000;
  } else {
    req.tv_sec=0; req.tv_nsec=us*1000;
  };
  nanosleep(&req, &rem);
}

size_t filesize(FILE *f)
{
  size_t s;
  fseek(f,0,SEEK_END); s=ftell(f); rewind(f);
  return s;
}

void lswap(void *x)
{
  int y;
  y=(*(int *)x & 0xff000000) >> 24;
  y|=(*(int *)x & 0x00ff0000) >> 8;
  y|=(*(int *)x & 0x0000ff00) << 8;
  y|=(*(int *)x & 0x000000ff) << 24;
  *(int*)x=y;
}

void sswap(short int *x)
{
  short int y;
  y=(*x & 0xff00) >> 8;
  y|=(*x & 0x00ff) << 8;
  *x=y;
}

unsigned char rxchar(int fd)
{
  unsigned char c=0xff;
  read(fd,&c,1);
  return c;
}

void txchar(int fd, unsigned char c)
{
  write(fd,&c,1);
}

int send_kernel_file(int fd, char *file)
{
  FILE *fp;
  int size, stat=0; 
  short h8size;
  unsigned char *buf;
  unsigned char *p, *q;
  unsigned char c;
  int i;
  fp=fopen(file,"rb");
  if (fp==NULL) {perror(file); return -1;};
  size=filesize(fp);
  buf=malloc(size);
  if (buf==NULL) {fclose(fp); perror("malloc"); return -1;};
  fread(buf, 1, size, fp);
  fclose(fp);
  h8size=size;
  sswap(&h8size);
  write(fd,&h8size,2);
  read(fd,&h8size,2);
  p=buf; q=buf;
  for (i=0; i<OVERLAP; i++) {
    write(fd,p,1); p++; size--;
  };
  while (size>0) {
    write(fd,p,1); read(fd,&c,1);
    if (c!=*q) {fprintf(stderr,"kernel send error %02X %02X\n",*p,c); stat=-1; break;};
    p++; q++; size--;
  };
  for (i=0; i<OVERLAP; i++) {
    read(fd,&c,1); 
    if (c!=*q) {fprintf(stderr,"kernel send error %02X %02X\n",*p,c); stat=-1; break;};
    q++;
  };
  free(buf);
  return stat;
}

int send_kernel_buf(int fd, const unsigned char *buf, int size)
{
  int stat=0; 
  short h8size;
  unsigned char *p, *q;
  unsigned char c;
  int i;
  h8size=size;
  sswap(&h8size);
  write(fd,&h8size,2);
  read(fd,&h8size,2);
  p=(char *)buf; q=p;
  for (i=0; i<OVERLAP; i++) {
    write(fd,p,1); p++; size--;
  };
  while (size>0) {
    write(fd,p,1); read(fd,&c,1);
    if (c!=*q) {fprintf(stderr,"kernel send error %02X %02X\n",*p,c); stat=-1; break;};
    p++; q++; size--;
  };
  for (i=0; i<OVERLAP; i++) {
    read(fd,&c,1); 
    if (c!=*q) {fprintf(stderr,"kernel send error %02X %02X\n",*p,c); stat=-1; break;};
    q++;
  };
  return stat;
}

int send_pgm(int fd, char *file)
{
  FILE *fp;
  int size, send;
  int addr, h8addr;
  int lastad, lastlen;
  int stat=0;
  unsigned short h8len;
  unsigned char *buf;
  unsigned char *p;
  unsigned char c;
  int i;
  fp=fopen(file,"rb");
  if (fp==NULL) {perror(file); return -1;};
  size=filesize(fp);
  buf=malloc(size);
  if (buf==NULL) {fclose(fp); perror("malloc"); return -1;};
  fread(buf, 1, size, fp);
  fclose(fp);
  lastad=-1; lastlen=-1;
  p=buf; addr=0; send=size;
  while (send>0) {
    i=BUFLEN; if (send<i) i=send;
    if (addr!=lastad) {
      h8addr=addr; lswap(&h8addr);
      txchar(fd,CMD_ADDR); write(fd,&h8addr,4);
      lastad=addr;
    };
    if (i!=lastlen) {
      h8len=i; sswap(&h8len);
      txchar(fd,CMD_LEN); write(fd,&h8len,2);
      lastlen=i;
    };
    txchar(fd,CMD_BUF); write(fd,p,i);
    txchar(fd,CMD_PGM);
    c=rxchar(fd);
    //printf("%d bytes written at %06X, status %d\n",i,addr,c);
    if (c!=0) {fprintf(stderr,"h8 flash write failed\n"); stat=-1; break;};
    addr+=i; lastad+=i; p+=i; send-=i;
  };
  #if 0
  p=buf; addr=0; send=size;
  while (send>0) {
    i=BUFLEN; if (send<i) i=send;
    if (addr!=lastad) {
      h8addr=addr; lswap(&h8addr);
      txchar(fd,CMD_ADDR); write(fd,&h8addr,4);
      lastad=addr;
    };
    if (i!=lastlen) {
      h8len=i; sswap(&h8len);
      txchar(fd,CMD_LEN); write(fd,&h8len,2);
      lastlen=i;
    };
    txchar(fd,CMD_BUF); write(fd,p,i);
    txchar(fd,CMD_VRFY);
    c=rxchar(fd);
    //printf("%d bytes verified at %06X, status %d\n",i,addr,c);
    if (c!=0) {fprintf(stderr,"h8 flash verify failed\n"); stat=-2; break;};
    addr+=i; lastad+=i; p+=i; send-=i;
  };
  #endif
  free(buf);
  if (stat==0) printf("%d bytes written\n",size);
  return stat;
}

void print_usage(char *n, char *d)
{
  printf("Usage: %s [options] h8pgmfile\n",n);
  printf("Options: -h            : this help\n"
	 "         -d device     : serial device (defaults to %s)\n",d);
  printf("         -k bootkernel : bootloader kernel to use (default builtin)\n");
}

int main(int argc, char *argv[])
{
  int i;
  unsigned char c;
  int new_baud;
  char *dev="/dev/ttyS1";
  char *bootname=NULL;
  char *pgmname="h8pgm.bin";

  /* analyze options */
  while ((i=getopt(argc,argv,"hd:k:"))>=0) {
    switch (i) { 
    case 'h':
	print_usage(argv[0],dev);
	return 0;
	break;
    case 'k':
    	bootname=optarg;
	break;
    case 'd':
	dev=optarg;
	break;
    }
  };
  if (optind<argc) pgmname=argv[optind];

  uart_fd=inituart(dev);
  if (uart_fd<=0) return 1;
  for (i=0; i<CONNTRY; i++) {
    if (select_fd(uart_fd)) {
      c=rxchar(uart_fd);
      if (c==0) break;
    };
    txchar(uart_fd,0);
    udelay(3000);
  };
  if (i>=CONNTRY) {fprintf(stderr,"connection to h8 failed\n"); return 1;};
  txchar(uart_fd,0x55);
  alarm(1);
  c=rxchar(uart_fd);
  alarm(0);
  if (c!=0xaa) {fprintf(stderr,"answer 0xAA from h8 failed\n"); return 1;};
  printf("send boot kernel\n");
  if (bootname) 
    i=send_kernel_file(uart_fd, bootname);
  else 
    i=send_kernel_buf(uart_fd,bootld,sizeof(bootld));
  if (i) return 1;
  c=0;
  printf("wait for erase status\n");
  for (i=0; i<WAITFORERASE; i++) {
    if (select_fd(uart_fd)) {
      c=rxchar(uart_fd);
      break;
    };
    udelay(10000);
  };
  if (c!=0xaa) {fprintf(stderr,"h8 flash erase failed\n"); return 1;};
  txchar(uart_fd,CMD_ID);
  alarm(1);
  if ((rxchar(uart_fd)!='m') || rxchar(uart_fd)!='g') {fprintf(stderr,"bootloader not identified\n"); return 1;};
  c=rxchar(uart_fd);
  alarm(0);
  new_baud=B9600;
  switch (c) {
    case 47 : new_baud=B115200; c=3; break;	// 16 MHz
    case 51 : new_baud=B38400; c=12; break;	// 14.7456 MHz
    default : printf("unknown SCI1_BRR %d, keep old baudrate\n",c); break;
  };
  txchar(uart_fd,CMD_BAUD); txchar(uart_fd,c); usleep(3000);
  setbaud(uart_fd,new_baud);

  txchar(uart_fd,CMD_ID);
  alarm(1);
  if ((rxchar(uart_fd)!='m') || rxchar(uart_fd)!='g') {fprintf(stderr,"new baudrate setting failed\n"); return 1;};
  c=rxchar(uart_fd);
  alarm(0);
  printf("send flash data\n");
  i=send_pgm(uart_fd, pgmname);
  if (i) return 1;
  return 0;
}
